"""
Basic views for Application
===========================
"""

import base64
import random
import logging
from typing import Optional
from datetime import datetime, timedelta
from flask import Blueprint, flash, redirect, render_template, request, url_for, jsonify
from .models import Item, User, Bid
from .auth import login_required, current_user

# Main blueprint.
bp = Blueprint('views', __name__)
api = Blueprint('api', __name__, url_prefix='/api')

logger = logging.getLogger(__name__)

# Blueprint for documentation.
docs_bp = Blueprint('docs', __name__)

MIN_BID_INCREMENT = 1


def get_winning_bid(item: Item) -> Optional[Bid]:
    """
    Return the (currently) winning bid for the given item.

    If there are no bids, or the item is not yet closed, return None.

    :param item: The item to get the winning bid for.
    :return: The winning bid, or None.
    """

    winning_bid = None
    try:
        winning_bid = Bid.objects(item=item) \
            .filter(created_at__lt=item.closes_at) \
            .order_by('-amount') \
            .first()
    except Exception as exc:
        logger.warning("Error getting winning bid: %s", exc, exc_info=True, extra={
            'item_id': item.id,
        })

    return winning_bid


def get_item_price(item: Item) -> int:
    """
    Return the current price of the given item.

    If there are no bids, return the starting bid.

    :param item: The item to get the price for.
    :return: The current price.
    """

    winning_bid = get_winning_bid(item)
    if winning_bid:
        return winning_bid.amount + MIN_BID_INCREMENT
    else:
        return item.starting_bid


@bp.route("/")
def index() -> str:
    """
    Index page.

    """

    # Render template file. Template file is using Jinja2 syntax, and can be passed an arbitrary number
    # of arguments to be used in the template. The template file is located in the templates directory.
    html = render_template("index.html.j2", title="TJTS5901 Example. It was changed by Arno.")
    return html

@bp.route("/test")
def test_item_adding():
    """
    Test item is added
    """
    # for test purpose create random number (unique mail address required)
    ran_int = str(random.randint(0,10000))

    # Create a new user
    user = User()
    user.email = "test" + ran_int + "@gmail.com"
    user.password = "placeholder"
    user.save()

    #load test image
    with open("src/tjts5901/test.jpeg", "rb") as image_file:
        image = image_file.read()

    # Create a new item
    item = Item()
    item.title = "Test title"
    item.description = "This is a test description"
    item.starting_bid = 100
    item.seller = user
    item.image = image
    item.save()

    return "OK"

@bp.route("/addItem", methods=('GET', 'POST'))
@login_required
def add_item():
    """
    AddItem page.

    """
    if request.method == 'POST':
        nItem = request.form.get('nItem')
        description = request.form.get('description')
        starting_price = int(request.form.get('sPrice'))
        image = request.files['image']
        error = None
        if not nItem:
            error = 'Item name is required.'
        if not starting_price or starting_price < 1:
            error = 'Starting price must be greater than 0.'

        if error is None:
            # Placeholder for user as long as login function is not defined
            ran_int = str(random.randint(0,10000))
            user = User()
            user.email = "test" + ran_int + "@gmail.com"
            user.password = "placeholder"
            user.save()
            try:
                item = Item(
                    title=nItem,
                    description=description,
                    starting_bid=starting_price,
                    seller=user,
                    closes_at=datetime.utcnow() + timedelta(days=1),
                    image=image

                )
                item.save()

            except Exception as exc:
                error = f"Error creating item: {exc!s}"
                print('error:', exc)
            else:
                return redirect(url_for('views.list_bid'))

            flash(error)

    item = render_template("addItem.html")
    return item


@bp.route("/listBid")
@login_required
def list_bid():
    """
    page that lists all items currently on auction

    """
    # only on sale items are shown
    items = Item.objects.filter(closes_at__gt=datetime.utcnow()) \
        .order_by('-closes_at') 

    for item in items:
        item.image_base64 = base64.b64encode(item.image.read()).decode('utf-8')
        item.current_price = get_item_price(item) - MIN_BID_INCREMENT

    html = render_template("listBid.html", items=items)
    return html


@bp.route("item/<id>/bid", methods=('POST',))
@login_required
def bid(id):
    """
    method that saves bid on object

    If the bid is valid, create a new bid and redirect to the item view page.
    Otherwise, display an error message and redirect back to the item view page.

    """
    # frontend not implemented yet, therefore not active
    item = Item.objects.get_or_404(id=id)
    min_amount = item.starting_bid

    amount = int(request.form.get('bid'))

    if amount <= min_amount:
        flash(("Bid must be at least %(min_amount)s", min_amount))
        return redirect(url_for('views.list_bid'))

    if item.closes_at < datetime.utcnow():
        flash("This item is no longer on sale.")
        return redirect(url_for('views.list_bid'))

    try:
        # Notice: if you have integrated the flask-login extension, use current_user
        # instead of g.user
        bid = Bid(
            item=item,
            bidder=current_user,
            amount=amount,
        )
        bid.save()
    except Exception as exc:
        flash(_("Error placing bid: %(exc)s", exc=exc))
    else:
        flash(("Bid placed successfully!"))

    return redirect(url_for('views.list_bid', id=id))
    

@api.route('<id>/bids', methods=('GET',))
@login_required
def api_item_bids(id):
    """
    Get the bids for an item.

    :param id: The id of the item to get bids for.
    :return: A JSON response containing the bids.
    """

    item = Item.objects.get_or_404(id=id)
    bids = []
    for bid in Bid.objects(item=item).order_by('-amount'):
        bids.append(bid.to_json())

    return jsonify({
        'success': True,
        'bids': bids
    })


@api.route('<id>/bids', methods=('POST',))
@login_required
def api_item_place_bid(id):
    """
    Place a bid on an item.

    If the bid is valid, create a new bid and return the bid.
    Otherwise, return an error message.
    
    Only accepts `REF_CURRENCY` bids.

    :param id: The id of the item to bid on.
    :return: A JSON response containing the bid.
    """
    
    item = Item.objects.get_or_404(id=id)
    min_amount = get_item_price(item)

    try:
        amount = int(request.form['amount'])
    except KeyError:
        return jsonify({
            'success': False,
            'error': ("Missing required argument %(argname)s", argname:='amount')
             })
    except ValueError:
        return jsonify({
            'success': False,
            'error': ("Invalid value for argument %(argname)s", argname:='amount')
        })
    except Exception as exc:
        return jsonify({
            'success': False,
            'error': ("Error parsing argument %(argname)s: %(exc)s", argname:='amount', exc:=exc)
        })

    if amount < min_amount:
        return jsonify({
            'success': False,
            'error': ("Bid must be at least %(min_amount)s", min_amount:=min_amount)
        })

    if item.closes_at < datetime.utcnow():
        return jsonify({
            'success': False,
            'error': ("This item is no longer on sale.")
        })

    try:
        bid = Bid(
            item=item,
            bidder=current_user,
            amount=amount,
        )
        bid.save()
    except Exception as exc:
        logger.error("Error placing bid: %s", exc, exc_info=True, extra={
            'item_id': item.id,
            'bidder_id': current_user.id,
            'amount': amount,
        })

        return jsonify({
            'success': False,
            'error': ("Error placing bid: %(exc)s", exc:=exc)
        })

    print(type(bid))
    return jsonify({
        'success': True,
        'bid': bid.to_json()
    })

@bp.route("/view/<id>")
@login_required
def page_bid(id):
    """
    Item view page.

    Displays the item details, and a form to place a bid.
    """

    item = Item.objects.get_or_404(id=id)

    # Set the current price for the bid according to current highest bid
    winning_bid = get_winning_bid(item)
    min_bid = get_item_price(item)
    current_price = min_bid - MIN_BID_INCREMENT

    item.image_base64 = base64.b64encode(item.image.read()).decode('utf-8')

    # Dark pattern to show enticing message to user
    #if item.closes_at < datetime.utcnow() + timedelta(hours=1):
    #    flash("This item is closing soon! Act now! Now! Now!")

    return render_template('view.html', item=item, current_price=current_price, min_bid=min_bid)

